local gameBoard = {}
gameBoard.width = 10
gameBoard.height = 20
gameBoard.tileSize = 0 -- This changes dynamically based on screen resolution
gameBoard.offsetX = 0
gameBoard.offsetY = 0
gameBoard.data = {}

local marginPercent = 0.2

function gameBoard:load(width, height)
    local marginX = width * marginPercent
    local marginY = height * marginPercent

    local availableWidth = width - 2 * marginX
    local availableHeight = height - 2 * marginY

    -- Find the minimum num of pixels each tile can be based on
    -- the dimensions of the screen after accounting for margins
    self.tileSize = math.min(availableWidth / self.width, availableHeight / self.height)

    -- Center the game board
    self.offsetX = (width - (self.width * self.tileSize)) / 2
    self.offsetY = (height - (self.height * self.tileSize)) / 2

    -- Init empty data layer where 0 = empty tile and 1 = filled tile
    for y = 1, self.height do
        self.data[y] = {}
        for x = 1, self.width do
            self.data[y][x] = { filled = false, color = nil }
        end
    end
end

function gameBoard:draw()
    -- First: draw the background grid
    love.graphics.setColor(1, 1, 1)

    for y = 1, self.height do
        for x = 1, self.width do
            -- Draw the grid lines. Reduce x and y offset by 1 for correct alignment
            love.graphics.rectangle(
                "line",
                self.offsetX + (x - 1) * self.tileSize,
                self.offsetY + (y - 1) * self.tileSize,
                self.tileSize,
                self.tileSize
            )

            -- Draw the locked Tetromino blocks if filled
            if self.data[y][x].filled then
                local drawX = self.offsetX + (x - 1) * self.tileSize
                local drawY = self.offsetY + (y - 1) * self.tileSize

                love.graphics.setColor(table.unpack(self.data[y][x].color))
                love.graphics.rectangle("fill", drawX, drawY, self.tileSize, self.tileSize)

                love.graphics.setColor(1, 1, 1)
            end
        end
    end
end

function gameBoard:clearRows()
    local clearable = {}

    -- Identify full rows starting from the bottom
    for y = self.height, 1, -1 do
        local isFull = true

        for x = 1, self.width do
            if not self.data[y][x].filled then
                isFull = false
                break
            end
        end

        if isFull then
            table.insert(clearable, y)
        end
    end

    local clearCount = 0
    -- Clear rows and shift everything down
    for _, rowIndex in ipairs(clearable) do
        local adjustedIndex = rowIndex + clearCount
        for i = adjustedIndex, 2, -1 do
            self.data[i] = self.data[i - 1]
        end

        -- Clear the top row after shifting everything down
        for j = 1, self.width do
            self.data[1][j] = { filled = false, color = nil }  -- Clear the top row correctly
        end

        clearCount = clearCount + 1
    end

    return #clearable
end


function gameBoard:print()
    for _, row in ipairs(self.data) do
        local rowString = ""
        for _, value in ipairs(row) do
            if value.filled then
                rowString = rowString .. "1 | "
            else
                rowString = rowString .. "0 | "
            end
        end
        print(rowString:sub(1, -4))  -- Remove the trailing " | "
    end
end


return gameBoard
