Tetromino = {}
Tetromino.__index = Tetromino

local L_BLOCK_BP = {
    { 1, 0 },
    { 1, 0 },
    { 1, 1 },
}

local INVERSE_L_BLOCK_BP = {
    { 0, 1 },
    { 0, 1 },
    { 1, 1 },
}

local Z_BLOCK_BP = {
    { 1, 1, 0 },
    { 0, 1, 1 },
}

local INVERSE_Z_BLOCK_BP = {
    { 0, 1, 1 },
    { 1, 1, 0 },
}

local T_BLOCK_BP = {
    { 0, 1, 0 },
    { 1, 1, 1 },
}

local LINE_PIECE_BP = {
    { 1 },
    { 1 },
    { 1 },
    { 1 }
}

local BLOCK_BP = {
    { 1, 1 },
    { 1, 1 }
}

local function getWidth(blueprint)
    local maxWidth = 0
    for _, row in ipairs(blueprint) do
        local rowWidth = #row
        if rowWidth > maxWidth then
            maxWidth = rowWidth
        end
    end
    return maxWidth
end

-- Tetrominos can be the following types:
-- "l" for "L block"
-- "z" for "Z block"
-- "t" for "T block"
-- "line" for "Line piece"
-- "block" for "Square block"
-- "li" for "Inverse L block"
-- "zi" for "Inverse Z block"

-- type could be "l" or "z" or "line" or "block" or "t"
function Tetromino:new(typeId, midpoint)
    local newTetromino = {
        x = midpoint,
        y = 1,
        moveTimer = 0,
        moveInterval = 0.075,
        fallingMoveTimer = 0,
        fallingInterval = 0.5,
        isLocked = false
    } -- based on game board coordinates

    local type = Tetromino:getTypeFromId(typeId)

    if type == "l" then
        newTetromino.blueprint = L_BLOCK_BP
        newTetromino.color = { 1, 0, 0 }
    elseif type == "z" then
        newTetromino.blueprint = Z_BLOCK_BP
        newTetromino.color = { 0, 1, 0 }
    elseif type == "t" then
        newTetromino.blueprint = T_BLOCK_BP
        newTetromino.color = { 0, 0, 1 }
    elseif type == "line" then
        newTetromino.blueprint = LINE_PIECE_BP
        newTetromino.color = { 1, 1, 0 }
    elseif type == "block" then
        newTetromino.blueprint = BLOCK_BP
        newTetromino.color = { 0, 1, 1 }
    elseif type == "li" then
        newTetromino.blueprint = INVERSE_L_BLOCK_BP
        newTetromino.color = { 1, 0.5, 0.5 }
    elseif type == "zi" then
        newTetromino.blueprint = INVERSE_Z_BLOCK_BP
        newTetromino.color = { 0.5, 1, 0.5 }
    end

    return setmetatable(newTetromino, self)
end

function Tetromino:getTypeFromId(id)
    if id == 1 then
        return "line"
    elseif id == 2 then
        return "block"
    elseif id == 3 then
        return "z"
    elseif id == 4 then
        return "zi"
    elseif id == 5 then
        return "l"
    elseif id == 6 then
        return "li"
    elseif id == 7 then
        return "t"
    end
end

function Tetromino:draw(gameBoard)
    love.graphics.setColor(table.unpack(self.color))

    for y, row in ipairs(self.blueprint) do
        for x, val in ipairs(row) do
            if val == 1 then
                -- Calculate the actual position on the screen
                local drawX = gameBoard.offsetX + (self.x + (x - 2)) * gameBoard.tileSize
                local drawY = gameBoard.offsetY + (self.y + (y - 2)) * gameBoard.tileSize

                love.graphics.rectangle("fill", drawX, drawY, gameBoard.tileSize, gameBoard.tileSize)
            end
        end
    end
end

local function canMoveY(tetromino, newX, newY, gameBoard)
    if tetromino.isLocked then return false end

    for y, row in ipairs(tetromino.blueprint) do
        for x, value in ipairs(row) do
            if value == 1 then
                local boardX = newX + x - 1
                local boardY = newY + y - 1

                -- Check if the position is outside the game board
                if boardY > gameBoard.height or boardY < 1 then
                    return false
                end

                -- Check if the position is already occupied
                if gameBoard.data[boardY][boardX].filled then
                    return false
                end
            end
        end
    end
    return true
end

local function canMoveX(tetromino, newX, newY, gameBoard)
    if tetromino.isLocked then return false end

    for y, row in ipairs(tetromino.blueprint) do
        for x, value in ipairs(row) do
            if value == 1 then
                local boardX = newX + x - 1
                local boardY = newY + y - 1

                -- Check if the position is outside the game board
                if boardX < 1 or boardX > gameBoard.width then
                    return false
                end

                -- Check if the position is already occupied
                if boardY >= 0 and gameBoard.data[boardY][boardX].filled then
                    return false
                end
            end
        end
    end
    return true
end


local function lockTetromino(tetromino, gameBoard)
    for y, row in ipairs(tetromino.blueprint) do
        for x, value in ipairs(row) do
            if value == 1 then
                local boardX = tetromino.x + x - 1
                local boardY = tetromino.y + y - 1

                print("[Locking] X: " .. boardX .. ", Y: " .. boardY)

                if boardX >= 1 and boardX <= gameBoard.width and boardY >= 1 and boardY <= gameBoard.height then
                    gameBoard.data[boardY][boardX] = { filled = true, color = tetromino.color }
                end
            end
        end
    end


    tetromino.isLocked = true
end

function Tetromino:update(dt, gameBoard, state)
    -- Handle falling movement
    self.fallingMoveTimer = self.fallingMoveTimer + dt

    local canMoveDown = self.fallingMoveTimer >= self.fallingInterval

    if canMoveDown then
        if canMoveY(self, self.x, (self.y + 1), gameBoard) then
            self.fallingMoveTimer = 0
            self.y = self.y + 1
        else
            lockTetromino(self, gameBoard)

            if self.y <= 1 then
                state:change("gameover")
            end

            return true
        end
    end

    return false
end

function Tetromino:move(dt, dir, gameBoard)
    self.moveTimer = self.moveTimer + dt

    local okTimer = self.moveTimer >= self.moveInterval

    if okTimer then
        if dir == -1 and canMoveX(self, self.x - 1, self.y, gameBoard) then
            self.moveTimer = 0
            self.x = self.x - 1
        elseif dir == 1 and canMoveX(self, self.x + 1, self.y, gameBoard) then
            self.moveTimer = 0
            self.x = self.x + 1
        end
    end
end

function Tetromino:increaseDropSpeed()
    self.fallingInterval = 0.025
end

function Tetromino:resetDropSpeed()
    self.fallingInterval = 0.5
end

function Tetromino:rotateClockwise(gameBoard)
    local transposed = {}
    local rows = #self.blueprint
    local cols = #self.blueprint[1]

    -- Handles the matrix rotation of pieces
    for i = 1, cols do
        transposed[i] = {}
        for j = 1, rows do
            transposed[i][j] = self.blueprint[rows - j + 1][i]
        end
    end

    local width = getWidth(self.blueprint)
    local newWidth = getWidth(transposed)

    -- Make sure max x bounds are respected
    if newWidth > width then
        while newWidth - 1 + self.x > gameBoard.width do
            self.x = self.x - 1
        end
    end

    -- Make sure min x bounds are respected
    if self.x < 0 then
        while self.x >= 0 do
            self.x = self.x + 1
        end
    end

    -- Make sure piece has totally empty spaces for rotation
    local canFit = true
    for y, row in ipairs(transposed) do
        for x, val in ipairs(row) do
            if val == 1 then
                local xPos = self.x + x - 1
                local yPos = self.y + y - 1

                if gameBoard.data[yPos][xPos].filled then
                    canFit = false
                    break
                end
            end
        end
    end

    if canFit then
        self.blueprint = transposed
    end
end

function Tetromino:rotateCounterClockwise(gameBoard)
    local transposed = {}
    local rows = #self.blueprint
    local cols = #self.blueprint[1]

    -- Handles the matrix rotation of pieces
    for i = 1, cols do
        transposed[i] = {}
        for j = 1, rows do
            transposed[i][j] = self.blueprint[j][cols - i + 1]
        end
    end

    local width = getWidth(self.blueprint)
    local newWidth = getWidth(transposed)

    -- Make sure max x bounds are respected
    if newWidth > width then
        while newWidth - 1 + self.x > gameBoard.width do
            self.x = self.x - 1
        end
    end

    -- Make sure min x bounds are respected
    if self.x < 0 then
        while self.x >= 0 do
            self.x = self.x + 1
        end
    end

    -- Make sure piece has totally empty spaces for rotation
    local canFit = true
    for y, row in ipairs(transposed) do
        for x, val in ipairs(row) do
            if val == 1 then
                local xPos = self.x + x - 1
                local yPos = self.y + y - 1

                if gameBoard.data[yPos][xPos].filled then
                    canFit = false
                    break
                end
            end
        end
    end

    if canFit then
        self.blueprint = transposed
    end
end

return Tetromino
